@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 
-# Copyright 2020-2021 Jean-Baptiste Delisle
+# Copyright 2020-2022 Jean-Baptiste Delisle
 #
 # This file is part of spleaf.
 #
@@ -26,6 +26,8 @@ __all__ = [
 ]
 
 import numpy as np
+from scipy.special import ive
+import warnings
 
 
 class Term:
@@ -877,6 +879,14 @@ class SumKernel(Kernel):
     for kernel in self._kernels:
       kernel._compute()
 
+  def _kernel_param(self, **kwargs):
+    raise NotImplementedError(
+      'The _kernel_param method should be implemented in SumKernel childs.')
+
+  def _kernel_param_back(self, *args):
+    raise NotImplementedError(
+      'The _kernel_param method should be implemented in SumKernel childs.')
+
   def _set_param(self, *args, **kwargs):
     for karg, arg in enumerate(args):
       par = self._param[karg]
@@ -984,6 +994,16 @@ class ProductKernel(Kernel):
         (k1 + 1) * self._kernel2._r] = self._kernel1._cov.phi[:, k1,
         None] * self._kernel2._cov.phi
 
+  def _kernel_param(self, **kwargs):
+    raise NotImplementedError(
+      'The _kernel_param method should be implemented in ProductKernel childs.'
+    )
+
+  def _kernel_param_back(self, *args):
+    raise NotImplementedError(
+      'The _kernel_param method should be implemented in ProductKernel childs.'
+    )
+
   def _set_param(self, *args, **kwargs):
     for karg, arg in enumerate(args):
       par = self._param[karg]
@@ -1617,52 +1637,60 @@ class _ESP_PKernel(SumKernel):
   This is not a valid kernel on its own.
   """
 
-  def __init__(self, P, eta):
+  def __init__(self, P, eta, nharm):
     self._P = P
     self._eta = eta
-    a0, a1, a2, nu = self._getcoefs()
-    self._exp = ExponentialKernel(a0, 0)
-    self._qp1 = QuasiperiodicKernel(a1, 0, 0, nu)
-    self._qp2 = QuasiperiodicKernel(a2, 0, 0, 2 * nu)
-    super().__init__(self._exp, self._qp1, self._qp2)
+    self._nharm = nharm
+    self._bessel = np.empty(self._nharm + 1)
+    self._calccoefs()
+    kernels = [ExponentialKernel(self._a[0], 0)] + [
+      QuasiperiodicKernel(self._a[k], 0, 0, k * self._nu)
+      for k in range(1, nharm + 1)
+    ]
+    super().__init__(*kernels)
     self._param = ['P', 'eta']
 
-  def _getcoefs(self):
+  def _calccoefs(self):
     self._eta2 = self._eta * self._eta
     self._f = 1 / (4 * self._eta2)
-    self._f2 = self._f * self._f
-    self._f2_4 = self._f2 / 4
-    self._deno = 1 + self._f + self._f2_4
-    a0 = 1 / self._deno
-    a1 = self._f * a0
-    a2 = self._f2_4 * a0
-    nu = 2 * np.pi / self._P
-    return (a0, a1, a2, nu)
+    # Compute one more Bessel function for derivative (see _kernel_param_back)
+    self._bessel = ive(np.arange(self._nharm + 2), self._f)
+    self._a = self._bessel[:-1].copy()
+    self._a[0] /= 2
+    self._deno = np.sum(self._a)
+    self._a /= self._deno
+    self._nu = 2 * np.pi / self._P
 
   def _kernel_param(self, P=None, eta=None):
     if P is not None:
       self._P = P
     if eta is not None:
       self._eta = eta
-    a0, a1, a2, nu = self._getcoefs()
-    return (dict(a=a0), dict(a=a1, nu=nu), dict(a=a2, nu=2 * nu))
+    self._calccoefs()
+    return ([dict(a=self._a[0])] +
+      [dict(a=self._a[k], nu=k * self._nu) for k in range(1, self._nharm + 1)])
 
-  def _kernel_param_back(self, gradExp, gradQP1, gradQP2):
+  def _kernel_param_back(self, *gradkernels):
     grad = {}
-    aga0 = self._exp._a * gradExp['a']
-    aga1 = self._qp1._a * gradQP1['a']
-    aga2 = self._qp2._a * gradQP2['a']
-    sgs = 2 * (aga0 + aga1 + aga2)
-    grad['P'] = -(self._qp1._nu * gradQP1['nu'] +
-      self._qp2._nu * gradQP2['nu']) / self._P
-    grad['eta'] = (sgs * (self._f + self._f2 / 2) / self._deno - 2 *
-      (aga1 + 2 * aga2)) / self._eta
+    grad['P'] = -np.sum([
+      self._kernels[k]._nu * gradkernels[k]['nu']
+      for k in range(1, self._nharm + 1)
+    ]) / self._P
+
+    ga = np.array([gk['a'] for gk in gradkernels])
+    # Use recurrence relations for derivative of Bessel functions:
+    da_df = self._bessel[1:] / 2
+    da_df[1:] += self._bessel[:-2] / 2
+    ddeno_df = np.sum(da_df)
+    da_df = (da_df - self._a * ddeno_df) / self._deno
+    df_deta = -2 * self._f / self._eta
+    grad['eta'] = df_deta * (da_df @ ga)
     return (grad)
 
 
 class ESPKernel(ProductKernel):
   r"""
-  The Exponential-sine periodic (ESP) kernel is a rank 15
+  The Exponential-sine periodic (ESP) kernel is a
   twice mean-square differentiable kernel
   approximating the squared-exponential periodic (SEP) kernel.
 
@@ -1674,12 +1702,16 @@ class ESPKernel(ProductKernel):
   while the ESP kernel approximates it with
 
   .. math:: k(\Delta t) = \sigma^2 k_\mathrm{ES}(\rho;\Delta t)
-      \frac{1 + f\cos\left(\nu \Delta t\right)
-      + \frac{f^2}{4} \cos\left(2\nu\Delta t\right)}{1+f+\frac{f^2}{4}},
+      \frac{I_0(f) + 2 \sum_{h=1}^{n_\mathrm{harm}}
+        I_h(f) \cos\left(h \nu \Delta t\right)}{
+        I_0(f) + 2 \sum_{h=1}^{n_\mathrm{harm}} I_h(f)},
 
   where :math:`\nu = \frac{2\pi}{P}`, :math:`f = \frac{1}{4\eta^2}`,
+  :math:`I_h` are the modified Bessel functions of the first kind,
   and :math:`k_\mathrm{ES}` is an :class:`ESKernel` with scale :math:`\rho`.
 
+  The rank of the ESP kernel is :math:`6 n_\mathrm{harm} + 3`.
+
   Parameters
   ----------
   sig : float
@@ -1690,15 +1722,16 @@ class ESPKernel(ProductKernel):
     Scale.
   eta : float
     Scale of oscillations.
+  nharm : int
+    Number of harmonics to include (2 by default).
   """
 
-
-  def __init__(self, sig, P, rho, eta):
+  def __init__(self, sig, P, rho, eta, nharm=2):
     self._sig = sig
     self._P = P
     self._rho = rho
     self._eta = eta
-    super().__init__(ESKernel(sig, rho), _ESP_PKernel(P, eta))
+    super().__init__(ESKernel(sig, rho), _ESP_PKernel(P, eta, nharm))
     self._param = ['sig', 'P', 'rho', 'eta']
 
   def _kernel_param(self, sig=None, P=None, rho=None, eta=None):
@@ -1870,7 +1903,7 @@ class MultiSeriesKernel(Kernel):
 
     return (grad)
 
-  def set_conditionnal_coef(self, alpha=1, beta=0, series_id=None):
+  def set_conditional_coef(self, alpha=1, beta=0, series_id=None):
     r"""
     Set the coefficients used for the conditional computations.
 
@@ -1894,6 +1927,16 @@ class MultiSeriesKernel(Kernel):
       self._cond_alpha = None
       self._cond_beta = None
 
+  def set_conditionnal_coef(self, *args, **kwargs):
+    r"""
+    Same as the :func:`set_conditional_coef` method (here for backward-compatibility).
+    """
+
+    warnings.warn(
+      "Please use the set_conditional_coef method (with a single n).",
+      DeprecationWarning)
+    self.set_conditional_coef(*args, **kwargs)
+
   def _compute_t2(self, t2, dt2, U2, V2, phi2, ref2left, dt2left, dt2right,
     phi2left, phi2right):
 
